@@ -65,6 +65,7 @@ class InfoBlox(requests.Session):
         dns_view=None,
         alias_types=None,
         log_change=False,
+        new_zone_fields=None,
         log=None,
     ):
         super(InfoBlox, self).__init__()
@@ -75,6 +76,7 @@ class InfoBlox(requests.Session):
         self.verify = verify
         self.apiver = apiver or '1.0'
         self.log_change = log_change
+        self.new_zone_fields = new_zone_fields or {}
         self.log = log
         if not apiver:
             self.apiver = self.get_api_version()
@@ -84,7 +86,7 @@ class InfoBlox(requests.Session):
 
     def request(self, method, url, **kwargs):
         if self.log_change and method not in ('GET', 'HEAD'):
-            print(f'{method} {url} {kwargs}')
+            self.log.info(f'{method} {url} {kwargs}')
         ret = super().request(method, self.url(url), **kwargs)
         try:
             ret.raise_for_status()
@@ -105,16 +107,48 @@ class InfoBlox(requests.Session):
         vers = ([int(i) for i in v.split('.')] for v in vers)
         return '.'.join(str(i) for i in sorted(vers)[-1])
 
+    def get_zone_fqdn(self, zone):
+        if zone.endswith('in-addr.arpa.'):
+            return '{0}/{1}'.format(
+                ".".join((["0"] * 4 + zone.split("."))[-4:-8:-1]),
+                (zone.count(".") - 2) * 8,
+            )
+        elif zone.endswith('ip6.arpa.'):
+            return '{0}/{1}'.format(
+                ":".join(
+                    map(
+                        ''.join,
+                        zip(*[iter((["0"] * 32 + zone.split("."))[-4:-36:-1])] * 4),
+                    )
+                ),
+                (zone.count(".") - 2) * 4,
+            )
+        else:
+            return zone[:-1]
+
     def get_zone(self, zone):
         return self.get(
             'zone_auth',
             params={
-                'fqdn': zone.rstrip('.'),
+                'fqdn': self.get_zone_fqdn(zone),
                 '_return_fields+': 'soa_default_ttl',
                 **({'view': self.dns_view} if self.dns_view else {}),
             },
         ).json()
 
+    def add_zone(self, zone):
+        fqdn = self.get_zone_fqdn(zone)
+        zone_format = 'IPV6' if ':' in fqdn else 'IPV4' if '/' in fqdn else 'FORWARDING'
+        return self.post(
+            'zone_auth',
+            json={
+                'fqdn': fqdn,
+                'zone_format': zone_format,
+                '_return_fields+': 'soa_default_ttl',
+                **self.new_zone_fields,
+            },
+        ).json()
+
     def get_records(self, type, fields, zone, default_ttl, **extra):
         ret = self.get(
             'record:{0}'.format(type.lower()),
@@ -122,7 +156,7 @@ class InfoBlox(requests.Session):
                 'zone': zone.rstrip('.'),
                 **extra,
                 '_return_fields+': ','.join(
-                    (() if type == 'NS' else ('ttl', 'use_ttl')) + fields
+                    (() if type == 'NS' else ('ttl', 'use_ttl')) + fields + ('name',)
                 ),
                 '_paging': 1,
                 '_max_results': 1000,
@@ -207,7 +241,7 @@ class InfoBloxProvider(BaseProvider):
     def __init__(
         self,
         id,
-        gridmaster,
+        endpoint,
         username,
         password,
         verify=True,
@@ -215,12 +249,14 @@ class InfoBloxProvider(BaseProvider):
         dns_view=None,
         alias_types=None,
         log_change=False,
+        create_zones=False,
+        new_zone_fields=None,
         *args,
         **kwargs,
     ):
         self.log = logging.getLogger(f'{self.__class__.__name__}[{id}]')
         self.conn = InfoBlox(
-            gridmaster,
+            endpoint,
             username,
             password,
             verify,
@@ -228,10 +264,12 @@ class InfoBloxProvider(BaseProvider):
             dns_view,
             alias_types,
             log_change,
+            new_zone_fields,
             self.log,
         )
+        self.create_zones = create_zones
         self.log.debug(
-            f'__init__: https://{username}@{gridmaster}/wapi/v{self.conn.apiver}/'
+            f'__init__: https://{username}@{endpoint}/wapi/v{self.conn.apiver}/'
         )
         super(InfoBloxProvider, self).__init__(id, *args, **kwargs)
 
@@ -241,7 +279,7 @@ class InfoBloxProvider(BaseProvider):
         supported_objects = self.conn.get('?_schema').json()['supported_objects']
         return {t for t in type_map if f'record:{t.lower()}' in supported_objects}
 
-    def _data_for(self, type, zone, default_ttl):
+    def _data_for(self, type, zone, default_ttl, target):
         spec = type_map[type]
         single_field = isinstance(spec, str)
         fields = (spec,) if single_field else (*spec,)
@@ -252,9 +290,9 @@ class InfoBloxProvider(BaseProvider):
                 name,
                 source,
                 (
-                    values[0][spec[0]]
-                    if {v[spec[1]] for v in values} == self.conn.alias_types
-                    else 'example.com.'
+                    values[0][spec[0]] + 'invalid.'
+                    if target and {v[spec[1]] for v in values} != self.conn.alias_types
+                    else values[0][spec[0]]
                 )
                 if type == 'ALIAS'
                 else values[0][spec]
@@ -275,16 +313,14 @@ class InfoBloxProvider(BaseProvider):
         zone_data = self.conn.get_zone(zone.name)
 
         if not zone_data:
-            if target:
-                raise ValueError(
-                    "Zone does not exist in InfoBlox: {0}".format(zone.name)
-                )
+            if target and not self.create_zones:
+                raise ValueError(f'Zone does not exist in InfoBlox: {zone.name}')
             return False
 
         default_ttl = zone_data[0]['soa_default_ttl']
 
         for type in sorted(self.SUPPORTS):
-            for t, n, s, v in self._data_for(type, zone.name, default_ttl):
+            for t, n, s, v in self._data_for(type, zone.name, default_ttl, target):
                 record_name = zone.hostname_from_fqdn(n)
                 record = Record.new(
                     zone,
@@ -310,7 +346,7 @@ class InfoBloxProvider(BaseProvider):
             if type == 'ALIAS':
                 spec = type_map[type]
                 for t in self.conn.alias_types:
-                    v = {spec[0]: value, spec[1]: t}
+                    v = {spec[0]: value[:-1], spec[1]: t}
                     self.conn.add_record(type, zone, new.name, v, new.ttl, default_ttl)
             else:
                 self.conn.add_record(type, zone, new.name, value, new.ttl, default_ttl)
@@ -330,11 +366,11 @@ class InfoBloxProvider(BaseProvider):
                 spec = type_map[type]
                 refs = {r[spec[1]]: r for r in ext.refs}
                 for t in self.conn.alias_types - {*refs}:
-                    v = {spec[0]: value, spec[1]: t}
+                    v = {spec[0]: value[:-1], spec[1]: t}
                     self.conn.add_record(type, zone, new.name, v, new.ttl, default_ttl)
                 for t in self.conn.alias_types & {*refs}:
                     if refs[t][spec[0]] != value:
-                        v = {spec[0]: value, spec[1]: t}
+                        v = {spec[0]: value[:-1], spec[1]: t}
                         self.conn.mod_record(type, refs[t], v, new.ttl, default_ttl)
                 self.conn.del_record(
                     r for t, r in refs.items() if t not in self.conn.alias_types
@@ -359,15 +395,17 @@ class InfoBloxProvider(BaseProvider):
 
     def _apply(self, plan):
 
-        zone = plan.desired.name[:-1]
+        zone = plan.desired.name
 
         zone_data = self.conn.get_zone(zone)
 
         if not zone_data:
-            raise ValueError("Zone does not exist in InfoBlox: {0}".format(zone.name))
+            if not self.create_zones:
+                raise ValueError(f'Zone does not exist in InfoBlox: {zone}')
+            zone_data = [self.conn.add_zone(zone)]
 
-        default_ttl = zone_data[0]['soa_default_ttl']
+        default_ttl = zone_data[0].get('soa_default_ttl', 3600)
 
         for change in plan.changes:
             class_name = change.__class__.__name__
-            getattr(self, '_apply_{}'.format(class_name))(zone, change, default_ttl)
+            getattr(self, f'_apply_{class_name}')(zone[:-1], change, default_ttl)
